from django.core.management.base import BaseCommand

from django.conf import settings
from django.contrib.auth.models import User

from gamoto import users, openvpn, iptables

import os
import sys


class Command(BaseCommand):
    help = 'OpenVPN client-connect scripts'

    def __init__(self, *a, **kw):
        BaseCommand.__init__(self, *a, **kw)

        self.ipt = iptables.IPTables()

    def add_arguments(self, parser):
        parser.add_argument('client_cmd_path', nargs='?', type=str)

    def connect(self, user):
        u = User.objects.get(username=user)

        if not u.is_active:
            self.stderr.write('User '+user+' disabled\n')
            sys.exit(1)

        if settings.MANAGE_IPTABLES:
            self.ipt.flushClient(user)
            subnets = openvpn.getRoutes(user)

            if subnets:
                ip = os.getenv('ifconfig_pool_remote_ip')
                for subnet in subnets:
                    self.ipt.allowClient(user, ip, subnet)

    def disconnect(self, user):
        if settings.MANAGE_IPTABLES:
            self.ipt.flushClient(user)

    def handle(self, *args, **options):
        script_type = os.getenv('script_type')
        user = os.getenv('common_name')
        if not user:
            self.stderr.write(
                'Error, no CN found. This script must be run by OpenVPN!')
            sys.exit(1)

        if settings.MANAGE_IPTABLES:
            self.ipt.setupIptables()

        if script_type == 'client-connect':
            self.connect(user)

        elif script_type == 'client-disconnect':
            self.disconnect(user)

        else:
            self.stderr.write('Not sure what you want me to do :(')
            sys.exit(1)



from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import Permission
from django.db.utils import IntegrityError
from django.conf import settings

from gamoto import ca, users

import os
import sys


class Command(BaseCommand):
    help = 'Configure the CA'

    def _status(self, var):
        if var:
            self.stdout.write(self.style.SUCCESS('OK'))
        else:
            self.stdout.write(self.style.SUCCESS('EXISTS'))

    def configureDB(self):
        # Setup DB
        self.stdout.write("Creating content types... ", ending="")
        self.stdout.flush()
        try:
            ctype = ContentType.objects.create(
                app_label='subnet',
                model='group'
            )
            ctype.save()
            self._status(True)
        except IntegrityError:
            self._status(False)

        self.stdout.write("Creating permissions... ", ending="")
        self.stdout.flush()
        try:
            ctype = ContentType.objects.get(app_label='auth', model='permission')
            default_perm = Permission.objects.create(
                content_type=ctype,
                name='Default Group',
                codename='default_group'
            )
            default_perm.save()
            self._status(True)
        except IntegrityError:
            self._status(False)

    def configureCA(self):
        # Ensure paths exist
        for path in [settings.BASE_PATH, settings.CA_PATH, settings.USER_PATH]:
            if not os.path.isdir(path):
                os.makedirs(path)

        # Setup CA
        myca = ca.CertificateAuthority(
            settings.CA_SETUP['ou'],
            settings.CA_SETUP['org'],
            settings.CA_SETUP['email'],
            settings.CA_SETUP['country'],
            settings.CA_SETUP['state'],
            settings.CA_SETUP['city'],
            key_path=settings.CA_PATH
        )

        self.stdout.write("Creating Diffie-Hellman parameters... ", ending="")
        self.stdout.flush()
        self._status(myca.createDH())

        self.stdout.write("Building certificate authority... ", ending="")
        self.stdout.flush()
        self._status(myca.createCA('openvpn'))

        self.stdout.write("Creating server certificate request... ", ending="")
        self.stdout.flush()
        self._status(myca.createCSR('openvpn'))

        self.stdout.write("Signing server certificate... ", ending="")
        self.stdout.flush()
        self._status(myca.signCSR('openvpn', server=True))

        self.stdout.write("Generating CRL... ", ending="")
        self.stdout.flush()
        self._status(myca.createCRL())

    def handle(self, *args, **options):
        if (users.getSystemUID() != os.getuid()):
            self.stdout.write(self.style.ERROR(
                "Please run this command as root, or " + settings.GAMOTO_USER))
            sys.exit(1)

        self.configureCA()
        self.configureDB()




from django.conf import settings

from gamoto import users


class IPTables(object):
    def __init__(self):
        self.tables = {
            'filter': {
                'INPUT': {'rules': [], 'default': None},
                'OUTPUT': {'rules': [], 'default': None},
                'FORWARD': {'rules': [], 'default': None}
            },
            'nat': {
                'PREROUTING': {'rules': [], 'default': None},
                'INPUT': {'rules': [], 'default': None},
                'OUTPUT': {'rules': [], 'default': None},
                'POSTROUTING': {'rules': [], 'default': None}
            }
        }

        self.refreshTables()

    def refreshTables(self):
        result, err = users.sudo('iptables-save')
        result = result.strip()

        if err:
            raise Exception(err)

        rules = [i for i in result.decode().split('\n')
                 if not i.startswith('#')]

        for r in rules:
            if r.startswith('*'):
                table = r.strip('*')
                if table not in self.tables:
                    self.tables[table] = {}
            elif r.startswith(':'):
                chain, default, b = r.strip(':').split()
                if default == '-':
                    default = None

                self.tables[table][chain] = {
                    'default': default,
                    'rules': []
                }
            elif r.startswith('-'):
                token, chain, args = r.split(None, 2)
                self.tables[table][chain]['rules'].append(args)

    def iptables(self, *rule):
        print(repr(rule))
        result, err = users.sudo('/sbin/iptables', *rule)
        if err:
            raise Exception(err)

    def getTable(self, table):
        return self.tables.get(table, {})

    def getChain(self, table, chain):
        return self.tables.get(table, {}).get(chain, {})

    def createChain(self, table, chain):
        if chain not in self.getTable(table):
            self.iptables('-t', table, '-N', chain)

    def addRule(self, table, chain, match, action, comment=None, top=False):
        rules = self.getChain(table, chain).get('rules', [])

        rtest = [match, '-j', action]
        if comment:
            rtest.append('-m comment --comment %s' % comment)

        if ' '.join(rtest) in rules:
            return

        rule = ['-t', table]

        if top:
            rule.extend(['-I', chain, '1'])
        else:
            rule.extend(['-A', chain])

        rule.extend(match.split())

        if comment:
            rule.extend(['-m', 'comment', '--comment', comment])

        rule.extend(['-j', action])

        self.iptables(*rule)

    def setupIptables(self):
        # Makes sure iptables has our chain
        self.refreshTables()

        self.createChain('filter', 'openvpn')

        input_rules = self.getChain('filter', 'INPUT').get('rules', [])

        self.addRule('filter', 'INPUT', '-i ' + settings.VPN_INTERFACE,
                     'openvpn', top=True)

        if settings.DEFAULT_DENY:
            self.addRule('filter', 'INPUT', '-i ' + settings.VPN_INTERFACE,
                         'DROP')

    def allowClient(self, name, src, dest):
        self.refreshTables()
        if not (name and src and dest):
            return None

        rule = '-i %s -s %s -d %s' % (settings.VPN_INTERFACE, src, dest)

        self.addRule('filter', 'openvpn', rule, 'ACCEPT', comment=name)

    def flushClient(self, name):
        self.refreshTables()
        rules = self.getChain('filter', 'openvpn').get('rules', [])

        for i, r in reversed(list(enumerate(rules))):
            if name in r:
                self.iptables('-t', 'filter', '-D', 'openvpn', '%s' % (i+1))



from django.test import TestCase, Client
from django.urls import reverse, exceptions
from django.urls.resolvers import URLPattern
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType

from bs4 import BeautifulSoup

from gamoto.tests import data

import os
import shutil


# Create your tests here.
class TestRockFace(TestCase):
    def setUp(self):
        from gamoto.management.commands import setup
        cmd = setup.Command()

        cmd._status = lambda var: None
        cmd.stdout.write = lambda var, ending=None: None

        cmd.configureDB()
        self.content_type = ContentType.objects.get(app_label='subnet')

        self.admin = User.objects.create_user(
            username='admin',
            password='12345',
            is_staff=True,
            is_superuser=True
        )
        self.user = User.objects.create_user(
            username='test',
            password='12345'
        )

        self.temp_path = os.path.join(os.getcwd(), '.test_tmp')
        os.makedirs(self.temp_path)
        os.makedirs(os.path.join(self.temp_path, 'users'))
        os.makedirs(os.path.join(self.temp_path, 'ccd'))
        os.makedirs(os.path.join(self.temp_path, 'ca'))

        self.test_settings = {
            'BASE_PATH': self.temp_path,
            'USER_PATH': os.path.join(self.temp_path, 'users'),
            'CA_PATH': os.path.join(self.temp_path, 'ca')
        }

    def tearDown(self):
        shutil.rmtree(self.temp_path)

    def test_authentication_required(self):
        from gamoto import urls

        redirect_map = []
        for url in urls.urlpatterns:
            if not isinstance(url, URLPattern):
                continue
            if url.name == "login":
                continue
            try:
                urlr = reverse(url.name)
                redirect_map.append((url.name, urlr))
            except exceptions.NoReverseMatch:
                pass

        for name, path in redirect_map:
            r = self.client.get(reverse(name))
            self.assertRedirects(r, reverse('login')+'?next=' + path)

    def test_login(self):
        self.client.login(username='test', password='12345')

    def formTester(self, name, redirects_to, data, kwargs={}):
        self.client.login(username='admin', password='12345')

        url = reverse(name, kwargs=kwargs)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        soup = BeautifulSoup(response.content, 'html.parser')

        form = soup.form

        self.assertEqual(form['action'], url)

        response = self.client.post(url, data)
        self.assertRedirects(response, reverse(redirects_to))

    def test_create_group(self):
        self.formTester('create_group', 'endpoints', {
            'name': 'test group',
            'default': False,
        })

        group = Group.objects.get(name='test group')

    def test_create_subnet(self):
        group = Group.objects.create(name='test group')
        group.save()

        subnets = ['192.168.0.1', '192.2.3.0/24', '10.10.10.10,10.20.0.0/16',
                   '10.10.0.0/24,10.20.0.0/16']

        for num, subnet in enumerate(subnets):
            self.formTester('add_group_subnet', 'endpoints', {
                'name': 'test subnet %s' % num,
                'subnet': subnet
            }, kwargs={'group_id': group.id})

        db_subnets = group.permissions.all()

        for subnet in db_subnets:
            self.assertIn(subnet.codename.split('_')[-1], subnets)

    def test_add_user_group(self):
        group = Group.objects.create(name='test group')
        group.save()
        perm = Permission.objects.create(
            name='test subnet',
            codename='network_10.1.1.0/24,10.1.2.0/24',
            content_type=self.content_type
        )
        perm.save()
        group.permissions.add(perm)
        group.save()

        with self.settings(**self.test_settings):
            user_groups = self.formTester('user_groups', 'users', {
                'groups': group.id,
            }, kwargs={'user_id': self.user.id})

        from gamoto import openvpn

        routes = openvpn.getRoutes('test')

        self.assertEquals(routes[0], '10.1.1.0/24')
        self.assertEquals(routes[1], '10.1.2.0/24')

    def test_default_group(self):
        with self.settings(**self.test_settings):
            self.formTester('create_group', 'endpoints', {
                'name': 'test group',
                'default': True,
            })

            group = Group.objects.get(name='test group')

            perm = Permission.objects.create(
                name='test subnet',
                codename='network_10.1.1.0/24,10.1.2.0/24',
                content_type=self.content_type
            )
            perm.save()
            group.permissions.add(perm)
            group.save()

            from gamoto import openvpn

            routes = openvpn.getRoutes('test')

            self.assertEquals(routes[0], '10.1.1.0/24')
            self.assertEquals(routes[1], '10.1.2.0/24')


class TestOpenvpn(TestCase):
    def setUp(self):
        self.sudo_commands = []
        self.blank_ipt = False

        self.ipt_copy = [i for i in data.IPTABLES_SAVE]

        from gamoto.management.commands import setup
        cmd = setup.Command()

        cmd._status = lambda var: None
        cmd.stdout.write = lambda var, ending=None: None

        cmd.configureDB()

        self.content_type = ContentType.objects.get(app_label='subnet')

        group = Group.objects.create(name="testgroup")

        perm = Permission.objects.create(
            codename="network_10.1.2.0/24",
            name="My network",
            content_type=self.content_type
        )

        group.permissions.add(perm)
        group.save()

        user = User.objects.create(username="test", email="test@mail.com")
        user.groups.add(group)
        user.save()

    def tearDown(self):
        data.IPTABLES_SAVE = self.ipt_copy

    def _fake_sudo(self, *a):
        for arg in a:
            if ' ' in arg:
                raise Exception(
                    "Space in '%s'. Are arguments split?" % repr(a))
        self.sudo_commands.append(' '.join(a))
        result = b''
        if 'iptables-save' in a:
            if self.blank_ipt:
                result = b'\n'
            else:
                result = bytes('\n'.join(data.IPTABLES_SAVE), encoding='ascii')

        elif a[1] == '-N':
            data.IPTABLES_SAVE.insert(5, ":%s - [0:0]" % a[2])

        elif a[1] == '-t':
            if a[2] == '-A':
                data.IPTABLES_SAVE.insert(6, ' '.join(a[3:]))

        elif a[1] == '-A':
            data.IPTABLES_SAVE.insert(6, ' '.join(a[1:]))

        return (result, b'')

    def _fake_get_user(self, name):
        return {
            'name': name,
            'uid': 1001,
            'gid': 1001,
            'gecos': name.capitalize,
            'home': '/var/lib/gamoto/users' + name,
            'shell': '/bin/false',
            'mfa': True
        }

    def test_blank_iptables(self):
        # For dumb reasons iptables-save can return nothing
        from gamoto.management.commands import openvpn
        from gamoto import users

        openvpn.users.sudo = self._fake_sudo
        openvpn.users.getUser = self._fake_get_user

        self.blank_ipt = True

        cmd = openvpn.Command()

        tables = cmd.ipt.refreshTables()

        cmd.ipt.setupIptables()

        tables = cmd.ipt.refreshTables()

        self.blank_ipt = False

        self.assertListEqual(self.sudo_commands, [
            'iptables-save', 'iptables-save', 'iptables-save',
            '/sbin/iptables -t filter -N openvpn',
            '/sbin/iptables -t filter -I INPUT 1 -i tun0 -j openvpn',
            '/sbin/iptables -t filter -A INPUT -i tun0 -j DROP',
            'iptables-save'
        ])

    def test_connect(self):
        from gamoto.management.commands import openvpn
        from gamoto import users

        openvpn.users.sudo = self._fake_sudo
        openvpn.users.getUser = self._fake_get_user

        cmd = openvpn.Command()
        os.environ.setdefault('common_name', 'test')
        os.environ.setdefault('script_type', 'client-connect')
        os.environ.setdefault('ifconfig_pool_remote_ip', '10.88.20.35')

        cmd.handle()

        self.assertListEqual(self.sudo_commands, [
            'iptables-save',
            'iptables-save',
            '/sbin/iptables -t filter -N openvpn',
            '/sbin/iptables -t filter -I INPUT 1 -i tun0 -j openvpn',
            '/sbin/iptables -t filter -A INPUT -i tun0 -j DROP',
            'iptables-save',
            '/sbin/iptables -t filter -A openvpn -i tun0 -s 10.88.20.35 -d'
            ' 10.1.2.0/24 -m comment --comment test -j ACCEPT'
        ])



from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.core import validators
from django import forms

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit


def validate_ipv4_cidr(value):
    """
    Validates a CIDR field
    """
    vals = value.split(',')

    for val in vals:
        if '/' in val:
            ip, prefix = val.split('/')
            try:
                assert(0 <= int(prefix) <= 32)
            except AssertionError:
                raise validators.ValidationError(
                    "CIDR prefix must be between 0 and 32")
            except ValueError:
                raise validators.ValidationError(
                    "CIDR prefix must be an integer")

            validators.validate_ipv4_address(ip)


class CIDRField(forms.GenericIPAddressField):
    def __init__(self, *a, **kwargs):
        self.default_validators = [validate_ipv4_cidr]
        forms.CharField.__init__(self, *a, **kwargs)


class GroupForm(forms.ModelForm):
    """
    Form for creating endpoint groups
    """

    default = forms.BooleanField(initial=False, required=False)

    class Meta:
        model = Group
        exclude = ()

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        permission_query = Permission.objects.filter(
            content_type=ContentType.objects.get(app_label='subnet')
        ).order_by('name')

        self.fields['permissions'].queryset = permission_query
        self.fields['permissions'].label_from_instance = self.label_permission

        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.add_input(Submit('submit', 'Save'))

    def label_permission(self, obj):
        subnet = obj.codename.split('_')[-1]
        return obj.name + ': ' + subnet


class UserGroupForm(forms.ModelForm):
    class Meta:
        model = User
        exclude = ('password', 'user_permissions', 'username', 'first_name',
                   'last_name', 'last_login', 'is_staff', 'is_active',
                   'is_superuser', 'email', 'date_joined')

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.add_input(Submit('submit', 'Save'))


class SubnetForm(forms.ModelForm):
    """
    Form for creating endpoint groups
    """
    class Meta:
        model = Permission
        exclude = ()

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.add_input(Submit('submit', 'Save'))


class GroupSubnetForm(forms.Form):
    """
    Add subnet to group form
    """
    name = forms.CharField(min_length=2, max_length=128)
    subnet = CIDRField()

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.add_input(Submit('submit', 'Save'))



from django.core.management.base import BaseCommand, CommandError
from django.db.utils import IntegrityError
from django.contrib.auth.models import User


class Command(BaseCommand):
    help = 'Set a user as admin'

    def add_arguments(self, parser):
        parser.add_argument('username', nargs=1, type=str)

    def handle(self, *args, **options):
        username = options['username'][0]
        u = User.objects.get(username=username)
        u.is_superuser = True
        u.is_staff = True
        u.save()
